@@ -7,7 +7,7 @@ from distutils.version import StrictVersion
 from os import chdir, get_terminal_size, path
 from pathlib import Path
 from queue import Empty
-from random import randint, sample
+from random import choice, randint, sample
 
 import requests
 import soco
@@ -1440,12 +1440,19 @@ def queue_item_core(speaker, action, args, type):
                 else:
                     position = int(current_position) + 1
             else:
-                error_and_exit(
-                    "Second parameter for '{}' must be 'next/play_next' or 'first/start'".format(
-                        action
+                try:
+                    position = int(args[1])
+                    position = position if position > 0 else 1
+                    position = position if position <= speaker.queue_size else 0
+                except ValueError:
+                    # Note that 'first/start' option is now redundant, but included
+                    # here for backward compatibility
+                    error_and_exit(
+                        "Second parameter for '{}' must be integer or 'next/play_next'".format(
+                            action
+                        )
                     )
-                )
-                return False
+                    return False
         # Select a random entry from the list, in case there's more than one
         item = items[randint(0, len(items) - 1)]
         print(speaker.add_to_queue(item, position=position))
@@ -1685,6 +1692,38 @@ def rename(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
+@one_or_two_parameters
+def add_uri_to_queue(speaker, action, args, soco_function, use_local_speaker_list):
+    uri = args[0]
+    position = 0
+    if len(args) == 2:
+        if args[1].lower() in ["first", "start"]:
+            position = 1
+        elif args[1].lower() in ["play_next", "next"]:
+            current_position = speaker.get_current_track_info()["playlist_position"]
+            if current_position == "NOT_IMPLEMENTED":
+                position = 1
+            else:
+                position = int(current_position) + 1
+        else:
+            try:
+                position = int(args[1])
+                position = position if position > 0 else 1
+                position = position if position <= speaker.queue_size else 0
+            except ValueError:
+                # Note that 'first/start' option is now redundant, but included
+                # here for backward compatibility
+                error_and_exit(
+                    "Second parameter for '{}' must be integer or 'next/play_next'".format(
+                        action
+                    )
+                )
+                return False
+
+    print(speaker.add_uri_to_queue(uri, position=position))
+    return True
+
+
 @one_parameter
 def play_file(speaker, action, args, soco_function, use_local_speaker_list):
     return play_local_file(speaker, args[0])
@@ -1696,9 +1735,15 @@ def play_m3u(speaker, action, args, soco_function, use_local_speaker_list):
     options = "" if len(args) == 1 else args[1]
     options = options.lower()
 
-    if not m3u_file.lower().endswith(".m3u"):
+    # Check for invalid options
+    invalid = set(options) - set("psr")
+    if invalid:
+        error_and_exit("Invalid option(s) '{}' supplied".format(invalid))
+        return False
+
+    if not (m3u_file.lower().endswith(".m3u") or m3u_file.lower().endswith(".m3u8")):
         error_and_exit(
-            "Filename '{}' does not end in '.m3u' or '.M3U'".format(m3u_file)
+            "Filename '{}' does not end in '.m3u' or '.m3u8'".format(m3u_file)
         )
         return False
 
@@ -1711,9 +1756,15 @@ def play_m3u(speaker, action, args, soco_function, use_local_speaker_list):
     if not tracks:
         error_and_exit("No tracks found in '{}'".format(m3u_file))
 
-    logging.info("Found {} tracks to play".format(len(tracks)))
+    logging.info("Found {} tracks".format(len(tracks)))
 
-    if "s" in options:
+    if "r" in options:
+        # Choose a single random track
+        track = choice(tracks)
+        tracks = [track]
+        logging.info("Choosing random track: {}".format(track.path))
+
+    elif "s" in options:
         logging.info("Shuffling playlist")
         # For some reason, 'shuffle(tracks)' does not work
         tracks = sample(tracks, len(tracks))
@@ -2019,4 +2070,5 @@ actions = {
     "play_local_file": SonosFunction(play_file, ""),
     "play_m3u": SonosFunction(play_m3u, ""),
     "play_local_m3u": SonosFunction(play_m3u, ""),
+    "add_uri_to_queue": SonosFunction(add_uri_to_queue, ""),
 }
