@@ -5,45 +5,59 @@ from http.server import HTTPServer
 from ipaddress import IPv4Address, IPv4Network
 from os import path
 from queue import Empty
+from socketserver import ThreadingMixIn
 from threading import Thread
 
 import ifaddr
 from RangeHTTPServer import RangeRequestHandler
 
-from .utils import error_and_exit, set_speaker_playing_local_file
+from .utils import error_and_exit, set_speaker_playing_local_file, set_sigterm
 
-# The port range to use
+# The HTTP server port range to use
 PORT_START = 54000
 PORT_END = 54099
 
-SUPPORTED_TYPES = ["MP3", "M4A", "MP4", "FLAC", "OGG", "WAV"]
+SUPPORTED_TYPES = ["MP3", "M4A", "MP4", "FLAC", "OGG", "WMA", "WAV", "AAC"]
+
+
+class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
+    """Handle requests in separate threads.
+
+    Use the MixIn approach instead of the core ThreadingHTTPServer
+    class for backwards compatibility with Python 3.5+
+    """
 
 
 class MyHTTPHandler(RangeRequestHandler):
-    filename = None
-    speaker_ip = None
+    def __init__(self, *args, filename=None, speaker_ips=None, **kwargs):
+        self.filename = filename
+        self.speaker_ips = speaker_ips
+        super().__init__(*args, **kwargs)
 
     def do_GET(self):
         logging.info("Get request received by HTTP server")
 
         # Only serve the specific file requested on the command line,
-        # and only to the specific Sonos speaker IP address
-        if (
-            MyHTTPHandler.filename != self.path.replace("/", "")
-            or self.client_address[0] != MyHTTPHandler.speaker_ip
-        ):
-            RangeRequestHandler.send_error(self, code=403, message="Access forbidden")
-            logging.info(
-                "Access to '{}' from '{}' forbidden".format(
-                    self.path, self.client_address[0]
-                )
+        # and only to Sonos speakers in the Sonos system
+        error = False
+        if self.path.replace("/", "") != self.filename:
+            logging.info("Access to file '{}' forbidden".format(self.path))
+            error = True
+        if self.client_address[0] not in self.speaker_ips:
+            logging.info("Access from IP '{}' forbidden".format(self.client_address[0]))
+            error = True
+        if error:
+            RangeRequestHandler.send_error(
+                self, code=403, message="SoCo-CLI HTTP Server: Access forbidden"
             )
             return
 
+        # Forward the GET request
         try:
             super().do_GET()
-        except:
-            # It's normal to hit exceptions with Sonos. Ignore.
+        except Exception as e:
+            # It's normal to hit some exceptions with Sonos.
+            logging.info("Exception ignored: {}".format(e))
             pass
 
     def log_message(self, format, *args):
@@ -51,23 +65,21 @@ class MyHTTPHandler(RangeRequestHandler):
         return
 
 
-def http_server(server_ip, directory, filename, speaker_ip):
+def http_server(server_ip, directory, filename, speaker_ips):
     # Set the directory from which to serve files, in the handler
-    handler = functools.partial(MyHTTPHandler, directory=directory)
+    # Set the specific filename and client IP that are authorised
+    handler = functools.partial(
+        MyHTTPHandler, filename=filename, speaker_ips=speaker_ips, directory=directory
+    )
 
-    # Set up the only filename that will be served, and the only IP to which
-    # it will be served
-    MyHTTPHandler.filename = filename
-    MyHTTPHandler.speaker_ip = speaker_ip
-
-    # Set up MIME types
+    # For possible future use: set up MIME types
     # MyHTTPHandler.extensions_map[".m4a"] = "audio/x-m4a"
     # MyHTTPHandler.extensions_map[".aac"] = "audio/aac"
 
     # Find an available port by trying ports in sequence
     for port in range(PORT_START, PORT_END + 1):
         try:
-            httpd = HTTPServer((server_ip, port), handler)
+            httpd = ThreadedHTTPServer((server_ip, port), handler)
             logging.info("Using {}:{} for web server".format(server_ip, port))
             httpd_thread = Thread(target=httpd.serve_forever, daemon=True)
             httpd_thread.start()
@@ -76,8 +88,8 @@ def http_server(server_ip, directory, filename, speaker_ip):
         except OSError:
             # Assume this means that the port is in use
             continue
-
-    return None
+    else:
+        return None
 
 
 def get_server_ip(speaker):
@@ -92,20 +104,41 @@ def get_server_ip(speaker):
                 )
                 if IPv4Address(speaker.ip_address) in network:
                     return ip.ip
-    return None
+    else:
+        return None
 
 
-def wait_until_stopped(speaker):
+def wait_until_stopped(speaker, aac_file=False):
     sub = speaker.avTransport.subscribe(auto_renew=True)
+    # Includes a hack for AAC files, which would be played in a repeat loop.
+    # The speaker never goes into a 'STOPPED' state, so we have
+    # to detect the shift into a 'TRANSITIONING' state after playback
+    has_played = False
     while True:
         try:
             event = sub.events.get(timeout=1.0)
-            if event.variables["transport_state"] not in ["PLAYING", "TRANSITIONING"]:
-                logging.info(
-                    "Speaker '{}' in state '{}'".format(
-                        speaker.player_name, event.variables["transport_state"]
-                    )
+            logging.info(
+                "Transport event: State = '{}'".format(
+                    event.variables["transport_state"]
                 )
+            )
+            # Special case for AAC files
+            if aac_file:
+                if event.variables["transport_state"] == "TRANSITIONING":
+                    logging.info("Transitioning event received")
+                    if has_played:
+                        logging.info("AAC: transition event indicating end of track")
+                        sub.unsubscribe()
+                        speaker.stop()
+                        return True
+                    else:
+                        logging.info("AAC: transition event before playback")
+                if event.variables["transport_state"] == "PLAYING":
+                    has_played = True
+                    logging.info("AAC: has_played set to True")
+            # General case for other file types. Note that pausing (PAUSED_PLAYBACK)
+            # does not terminate the loop.
+            if event.variables["transport_state"] == "STOPPED":
                 sub.unsubscribe()
                 return True
         except Empty:
@@ -148,7 +181,10 @@ def play_local_file(speaker, pathname):
     logging.info("Using server IP address: {}".format(server_ip))
 
     # Start the webserver (runs in a daemon thread)
-    httpd = http_server(server_ip, directory, url_filename, speaker.ip_address)
+    speaker_ips = []
+    for zone in speaker.all_zones:
+        speaker_ips.append(zone.ip_address)
+    httpd = http_server(server_ip, directory, url_filename, speaker_ips)
     if not httpd:
         error_and_exit("Cannot create HTTP server")
         return False
@@ -158,22 +194,36 @@ def play_local_file(speaker, pathname):
     logging.info("Stopping speaker '{}'".format(speaker.player_name))
     speaker.stop()
 
-    # Assemble the URI and send to the speaker for playback
+    # Assemble the URI
     uri = "http://" + server_ip + ":" + str(httpd.server_port) + "/" + url_filename
     logging.info("Playing file '{}' from directory '{}'".format(filename, directory))
     logging.info("Playback URI: {}".format(uri))
-    speaker.play_uri(uri)
+
+    # Send the URI to the speaker for playback
+    # A special hack is required for AAC files, which have to be treated like radio.
+    if filename.lower().endswith(".aac"):
+        aac_file = True
+        # speaker.play_uri(uri, force_radio=True)
+        speaker.play_uri(uri)
+    else:
+        aac_file = False
+        speaker.play_uri(uri)
 
     logging.info("Setting flag to stop playback on CTRL-C")
     set_speaker_playing_local_file(speaker)
 
     logging.info("Waiting for playback to stop")
-    wait_until_stopped(speaker)
+    # SIGTERM enablement experiment for AAC files only
+    if aac_file:
+        set_sigterm(True)
+    wait_until_stopped(speaker, aac_file=aac_file)
+    if aac_file:
+        set_sigterm(False)
 
     logging.info("Playback stopped ... terminating web server")
     httpd.shutdown()
 
-    logging.info("Web server thread terminated")
+    logging.info("Web server terminated")
     set_speaker_playing_local_file(None)
 
     return True
