@@ -154,6 +154,22 @@ class Speakers:
         os.remove(self.save_pathname)
         return self.save_pathname
 
+    def rename(self, old_name, new_name):
+        for index, speaker in enumerate(self._speakers):
+            if old_name.replace("’", "'") == speaker.speaker_name.replace("’", "'"):
+                # Update old record, delete, replace with new
+                new_speaker = speaker._replace(speaker_name=new_name)
+                del self._speakers[index]
+                self._speakers.append(new_speaker)
+                logging.info(
+                    "Renamed speaker in cache: '{}' to '{}'".format(old_name, new_name)
+                )
+                self.save()
+                logging.info("Saved updated cache file")
+                return True
+        logging.info("Failed to find speaker '{}' for rename".format(old_name))
+        return False
+
     @staticmethod
     def is_ipv4_address(ip_address):
         """Tests for an IPv4 address"""
@@ -171,7 +187,11 @@ class Speakers:
             for ip in adapter.ips:
                 if Speakers.is_ipv4_address(ip.ip):
                     network_ip = ipaddress.ip_network(ip.ip)
-                    if network_ip.is_private and not network_ip.is_loopback:
+                    if (
+                        network_ip.is_private
+                        and not network_ip.is_loopback
+                        and not network_ip.is_link_local
+                    ):
                         # Constrain the size of network that will be searched
                         netmask = ip.network_prefix
                         if netmask < self._min_netmask:
@@ -263,6 +283,7 @@ class Speakers:
         thread_list = []
         self._speakers = []
 
+        # THIS WILL STOP WORKING IN SOCO 0.21+
         # Disable SoCo caching to prevent problems caused by multiple households
         soco.core.zone_group_state_shared_cache.enabled = False
 
@@ -354,6 +375,18 @@ class Speakers:
                     return soco.SoCo(speaker.ip_address)
         return None
 
+    def get_all_speakers(self):
+        # THIS WILL STOP WORKING IN SOCO 0.21+
+        # Disable SoCo caching to prevent problems caused by multiple households
+        soco.core.zone_group_state_shared_cache.enabled = False
+        soco_speakers = []
+        for speaker in self._speakers:
+            soco_speakers.append(soco.SoCo(speaker.ip_address))
+        if soco_speakers:
+            return soco_speakers
+        else:
+            return None
+
     def print(self):
         if not self._speakers:
             return
